#include <linuxmt/types.h>
#include <linuxmt/sched.h>

/*
 *	This function can only be called with SS=DS=ES=kernel DS
 *	CS=kernel CS. SS:SP is the relevant kernel stack (IRQ's are
 *	take on 'current' kernel stack.
 *
 *	load_regs can also only be called from such a situation, thus
 *	we don't need to arse about with segment registers. The kernel isnt
 *	relocating.
 *
 *	To understand this you need to know how the compilers generate 8086
 *	stack frames. Functions normally start
 *
 *	push bp		! Save callers BP
 *	mov  bp,sp	! BP so we can use it to index registers
 *
 *	and end
 *
 *	mov sp,bp	! Fastest way to destroy local variables
 *	pop bp		! Restore callers BP
 *	ret		! Return address is top of stack now
 *
 *	save_regs() saves the callers registers and state then returns to
 *	the caller. It in effect freezes a copy of the caller context but doesnt
 *	prevent it being used temporarily beyond that as we do. 
 *
 *	load_regs() restores the callers context and returns skipping out of 
 *	schedule() [our the faked setup] back to the right place. 
 *
 *	fake_save_regs builds a stack frame that returns a new task to a
 *	kernel address of our choice using its own stack/context.
 *
 */

#asm
	.text
	.even
	.globl _save_regs
	
_save_regs:
!
!	Save the CPU registers
!	
	pop ax		! Return address save
	push bp		! schedule()'s bp
	pushf
	push di
	push si
	push bx
	push dx
!
!	Store KSP
!
	mov  bx,[_current]
	mov  [bx],sp
	push ax		! Return with the stack altered.
	ret

!
!	Unlike save_regs this doesn't return to the caller but
!	to the callers caller.
!	
	.globl _load_regs

_load_regs:
!
!	Recover KSP
!	
	mov bx,[_current]
	mov ax,[bx]
	mov sp,ax
!
!	Pull CPU state off the stack
!
	pop dx
	pop bx
	pop si
	pop di
	popf
!
!	Now we basically spoof the return code of schedule()
!
	pop bp		! BP of schedule()
	mov sp,bp	! As schedule() would do on its return to get SP
	pop bp		! Recover caller BP
	ret		! thus to caller of schedule()
	
!
!	System Call Vector
!
!	On entry we are on the wrong stack, DS, ES are wrong
!

	.globl _syscall
! This is from irqtab.c
	.extern stashed_ds

!
!	System calls enter here with bx as message pointer, ax as src/dest
!	and cx as function (fixed value in our case). syscall returns a value
!	in ax and modifies the message.
!
_syscall_int:
!
!	We know the process DS, we can discard it (indeed may change it)
!
	cli
	seg cs
	mov sc_tmp,ax
	seg cs
	mov stashed_ds,ax
	mov ds,ax
! Free an index register	
	mov si,bx
!
!	Find our TCB
!
	mov bx,_current
!
!	Stash user mode stack - needed for a reallocation case in
!	future.
!
	mov ax,sp
	mov 2[bx],ax
!
!	Finish switching to the right things
!
	mov [bx],ax
	mov sp,ax
	mov ax,ds
	mov es,ax
	mov ss,ax
	cld
!
!	Stack is now right, we can take interrupts OK
!
	sti
	seg	cs
	push	sc_tmp
	push	si
	push	cx
	push	dx
!
!	syscall(send/recv/both, block ptr, src/dst)
!
	call	_syscall
	pop	dx
	pop	cx
	pop	bx
	pop	bx
!
!	Now mend everything
!
	cli
	mov	dx,ax
	mov	bx,_current
	mov 	sp,2[bx]
	mov	ax,4[bx]
!
!	User segment recovery
!
	mov	ds,ax
	mov	es,ax
	mov	ss,ax
	sti
!
!	return with error info.
!
	mov	ax,dx
	iret
!
!	Done.
!
	.even	
sc_tmp:	.byte 0

	.globl _fake_save_regs
!
!	int used=fake_save_regs(sp,addr);
!
!	Build a fake return stack in kernel space so that
!	we can have a new task start at a chosen kernel 
!	function while on its kernel stack. We push the
!	registers suitably for
!
_fake_save_regs:
	push 	bp
	mov	bp,sp
	mov     bx,4[bp]
	mov	ax,6[bp]
!
!	Build a dummy stack return frame
!
	mov	[bx],ax		! Return address
	mov	ax,[bp]		! Caller BP
	mov	2[bx],ax	! Save caller BP
	mov	ax,bx		! Firstly get bx
	sub	ax,#2		! This is the apparent BP. It points
				! to caller BP and above it caller return
	mov	-4[bx],ax	! goes here.
!
!	Register State
!
	pushf
	pop	ax
	mov	-6[bx],ax	! Flags
	mov	-8[bx],di
	mov 	-10[bx],si
	mov	-12[bx],bx
	mov	-14[bx],dx
	mov	ax,#14
	pop	bp
	ret
#endasm

/*
 *	Make task t fork into kernel space. We are in kernel mode
 *	so we fork onto our kernel stack.
 */
 
void kfork_proc(t, addr)
struct task_struct *t;
char *addr;
{
	t->t_regs.ksp=&t->t_kstack[126];
	t->t_regs.ksp-=fake_save_regs(t->t_regs.ksp,addr);
	t->state=TASK_UNINTERRUPTIBLE;
	t->pid=0;
	t->ppid=0;
	t->uid=0;
	t->euid=0;
	t->suid=0;
	t->gid=0;
	t->egid=0;
	t->sgid=0;
	t->session=0;
	t->ppid=0;
	t->t_priority=10;
	t->t_count=0;
	wake_up_process(t);
	schedule();
}
